from ._neat import NEATClassifier
from ._neat import NEATRegressor
from ._wann import WANNClassifier
from ._wann import WANNRegressor


from ._version import __version__

__all__ = ['NEATClassifier', 'NEATRegressor', 'WANNClassifier', 'WANNRegressor',
           '__version__']
